/* Sudoku: the goal of this puzzle is to fill 81 1's into an array
   t[a][b][i][j][k][l], where a,b,i,j,k,l = 0..2, subject to the
   constraints that if t[a][b][i][j][k][l] =
   t[a'][b'][i'][j'][k'][l'], then:

   [a=a', b=b', i=i', j=j' => k=k', l=l'] and
   [a=a', b=b', i=i', k=k' => j=j', l=l'] and
   [a=a', b=b', j=j', l=l' => i=i', k=k'] and
   [i=i', j=j', k=k', l=l' => a=a', b=b'].

   forall (a,b,j,l) exists! (i,k)
   forall (a,b,i,k) exists! (j,l)
   forall (a,b,i,j) exists! (k,l)
   forall (i,j,k,l) exists! (a,b)
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "gettext.h"
#define _(String) gettext (String)

#include "sudoku.h"

static int debug = 0;

/* 0 means the field is available. 1 means it is taken, 2 means it is
   in conflict and cannot be taken any more */
typedef int hypercube_t[3][3][3][3][3][3];

/* define a permutation of indices */
typedef int perm_t[6];

static const perm_t permi[] = {
  { 0, 1, 2, 3, 4, 5 },
  { 0, 1, 2, 4, 3, 5 },
  { 0, 1, 4, 2, 5, 3 },
  { 4, 5, 0, 1, 2, 3 },
};
static const perm_t perm_id = 
  { 0, 1, 2, 3, 4, 5 };

/* holds a map from 9 to 3x3 */
struct shuffle_s {
  int x[9];
  int y[9];
};
typedef struct shuffle_s shuffle_t;

static const shuffle_t shuffle_id = {
  { 0, 0, 0, 1, 1, 1, 2, 2, 2, },
  { 0, 1, 2, 0, 1, 2, 0, 1, 2, }, 
};

/* returns a random shuffle */
static inline void random_shuffle(shuffle_t *sh) {
  int i;
  int r;
  int s[9];

  for (i=0; i<9; i++) {
    r = rand() % (i+1);
    s[i] = i;
    s[i] = s[r]; /* note: could have r==i */
    s[r] = i;
  }
  for (i=0; i<9; i++) {
    sh->x[i] = shuffle_id.x[s[i]];
    sh->y[i] = shuffle_id.y[s[i]];
  }
}

static inline int *get_position(hypercube_t t, int ix[6], const perm_t q) {
  return &t[ix[q[0]]][ix[q[1]]][ix[q[2]]][ix[q[3]]][ix[q[4]]][ix[q[5]]];
}

/* return 1 if a conflict was detected */
static inline int mark_macro(hypercube_t t, const perm_t q, int a, int b, int c, int d) {
  int ix[6];

  ix[0] = a;
  ix[1] = b;
  ix[2] = c;
  ix[3] = d;
  for (ix[4]=0; ix[4]<3; ix[4]++) {
    for (ix[5]=0; ix[5]<3; ix[5]++) {
      if (*get_position(t, ix, q) == 1) {
	return 1; /* conflict */
      }
      *get_position(t, ix, q) = 2;
    }
  }
  return 0;
}

/* mark the given position, and mark all arising new conflicts. Return
   1 if a pre-existing conflict was detected. */
static inline int mark(hypercube_t t, int ix[6], const perm_t q) {
  int r;

  r = mark_macro(t, permi[0], ix[q[0]], ix[q[1]], ix[q[2]], ix[q[3]]);
  if (r) {
    return 1;
  }
  r = mark_macro(t, permi[1], ix[q[0]], ix[q[1]], ix[q[2]], ix[q[4]]);
  if (r) {
    return 1;
  }
  r = mark_macro(t, permi[2], ix[q[0]], ix[q[1]], ix[q[3]], ix[q[5]]);
  if (r) {
    return 1;
  }
  r = mark_macro(t, permi[3], ix[q[2]], ix[q[3]], ix[q[4]], ix[q[5]]);
  if (r) {
    return 1;
  }
  *get_position(t, ix, q) = 1;
  return 0;
}

/* translate between types. Return -1 if inconsistent, else the number
   of non-zero members. */
static int sudoku_to_hypercube(sudoku_t su, hypercube_t t) {
  int i, j, k, l, a, b;
  int x, r;
  int ix[6];
  int count;

  count = 0;
  memset(t, 0, sizeof(hypercube_t));

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      for (k=0; k<3; k++) {
	for (l=0; l<3; l++) {
	  x = su[3*i+k][3*j+l];
	  if (x) {
	    count++;
	    a = (x-1) / 3;
	    b = (x-1) % 3;
	    ix[0] = a;
	    ix[1] = b;
	    ix[2] = i;
	    ix[3] = j;
	    ix[4] = k;
	    ix[5] = l;
	    r = mark(t, ix, perm_id);
	    if (r) {
	      return -1;
	    }
	  }
	}
      }
    }
  }
  return count;
}

/* translate hypercube to sudoku */
static void hypercube_to_sudoku(hypercube_t t, sudoku_t su) {
  int i, j, k, l, a, b;
  memset(su, 0, sizeof(sudoku_t));

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      for (k=0; k<3; k++) {
	for (l=0; l<3; l++) {
	  for (a=0; a<3; a++) {
	    for (b=0; b<3; b++) {
	      if (t[a][b][i][j][k][l] == 1) {
		su[3*i+k][3*j+l] = 3*a + b + 1;
		goto next;
	      }
	    }
	  }
	next:
	  ;
	}
      }
    }
  }
}

struct dim_s {
  int best[6];
  perm_t perm;
  int n;
};
typedef struct dim_s dim_t;

void sudoku_print(sudoku_t su, FILE *fout) {
  int i, j;
  
  for (i=0; i<9; i++) {
    if (i==0 || i==3 || i==6) {
      fprintf(fout, "+-------+-------+-------+\n");
    }
    for (j=0; j<9; j++) {
      if (j==0 || j==3 || j==6) {
	fprintf(fout, "| ");
      }
      if (su[i][j]) {
	fprintf(fout, "%d ", su[i][j]);
      } else {
	fprintf(fout, "  ");
      }
    }
    fprintf(fout, "|\n");
  }
  fprintf(fout, "+-------+-------+-------+\n");
}

void sudoku_print_big(sudoku_t su, FILE *fout) {
  int i, j;
  
  for (i=0; i<9; i++) {
    if (i==0 || i==3 || i==6) {
      fprintf(fout, "+---------+---------+---------+\n");
    } else {
      fprintf(fout, "|         |         |         |\n");
    }
    for (j=0; j<9; j++) {
      if (j==0 || j==3 || j==6) {
	fprintf(fout, "|");
      }
      if (su[i][j]) {
	fprintf(fout, " %d ", su[i][j]);
      } else {
	fprintf(fout, "   ");
      }
    }
    fprintf(fout, "|\n");
  }
  fprintf(fout, "+---------+---------+---------+\n");
}

void sudoku_print_small(sudoku_t su, FILE *fout) {
  int i, j;
  
  for (i=0; i<9; i++) {
    if (i==0 || i==3 || i==6) {
      fprintf(fout, "\n");
    }
    for (j=0; j<9; j++) {
      if (j==0 || j==3 || j==6) {
	fprintf(fout, " ");
      }
      if (su[i][j]) {
	fprintf(fout, "%d", su[i][j]);
      } else {
	fprintf(fout, " ");
      }
    }
    fprintf(fout, "\n");
  }
}

static inline void print_hypercube(hypercube_t t, FILE *fout) {
  sudoku_t su;

  hypercube_to_sudoku(t, su);
  sudoku_print(su, fout);
}

/* find plane parallel to the given one, with least zeros but no 1. */
static inline void best_plane(hypercube_t t, dim_t *b, const perm_t q) {
  int n, c;
  int ix[6];

  n = 10;
  for (ix[0]=0; ix[0]<3; ix[0]++) {
    for (ix[1]=0; ix[1]<3; ix[1]++) {
      for (ix[2]=0; ix[2]<3; ix[2]++) {
	for (ix[3]=0; ix[3]<3; ix[3]++) {
	  c=0;
	  for (ix[4]=0; ix[4]<3; ix[4]++) {
	    for (ix[5]=0; ix[5]<3; ix[5]++) {
	      if (*get_position(t, ix, q) == 0) {
		c++;
	      } else if (*get_position(t, ix, q) == 1) {
		goto next;
	      }
	    }
	  }
	  if (c<n) {
	    n = c;
	    memcpy(b->best, ix, sizeof(int[6]));
	    memcpy(b->perm, q, sizeof(perm_t));
	    b->n = n;
	  }
	next: 
	  ;
	}
      }
    }
  }
  if (debug > 1) {
    fprintf(stdout, "### %d -- %d %d %d %d %d %d -- %d %d %d %d %d %d\n", b->n, b->perm[0], b->perm[1], b->perm[2], b->perm[3], b->perm[4], b->perm[5], b->best[0], b->best[1], b->best[2], b->best[3], b->best[4], b->best[5]);
  }
}

/* Recurse through the solution tree. See sudoku_traverse for the
   meaning of mode. Counting of solutions starts at "count". */
static int traverse(hypercube_t t, int level, int mode, sudoku_callback_t *callback, int count, hypercube_t tr) {
  sudoku_t su;
  hypercube_t t1;
  dim_t dim;
  dim_t bestdim;
  perm_t q;
  int ix[6];
  int r;
  int i;
  shuffle_t sh;
  const shuffle_t *shp;
  int mode_mode = mode & MODE_MASK;

  if (debug) {
    fprintf(stdout, "Level %d:\n", level);
    print_hypercube(t, stdout);
  }
  if (level == 81) {
    count++;
    if (mode_mode==MODE_CALLBACK && callback != NULL) {
      hypercube_to_sudoku(t, su);
      callback->fun(su, callback->privdata);
    }
    if (mode_mode==MODE_FILL && tr != NULL) {
      memcpy(tr, t, sizeof(hypercube_t));
    }
    return count;
  }

  /* find the dimension with the least choices */
  bestdim.n=10;
  for (i=0; i<4; i++) {
    best_plane(t, &dim, permi[i]);
    if (dim.n < bestdim.n) {
      memcpy(&bestdim, &dim, sizeof(dim));
    }
  }
  
  if (debug > 1) {
    fprintf(stdout, "### bestdim.n = %d\n", bestdim.n);
  }

  if (bestdim.n==0) { /* inconsistent! */
    if (mode_mode==MODE_BRANCHES) {
      count++;
    }
    return count;
  }

  /* now iterate through the choices */
  memcpy(q, bestdim.perm, sizeof(perm_t));
  memcpy(ix, bestdim.best, sizeof(int[4]));
  if (mode & FLAG_RANDOM) {
    random_shuffle(&sh);
    shp = &sh;
  } else {
    shp = &shuffle_id;
  }  
    
  for (i=0; i<9; i++) {
    ix[4] = shp->x[i];
    ix[5] = shp->y[i];
    if (debug > 1) {
      fprintf(stdout, "### choice (%d %d %d %d %d %d) %d %d %d %d %d %d = %d\n", q[0], q[1], q[2], q[3], q[4], q[5], ix[0], ix[1], ix[2], ix[3], ix[4], ix[5], *get_position(t, ix, q));
    }
    if (*get_position(t, ix, q) == 0) {
      memcpy(t1, t, sizeof(hypercube_t));
      r = mark(t1, ix, q);
      if (!r) {
	count = traverse(t1, level+1, mode, callback, count, tr);
	if (mode_mode==MODE_UNIQUE && count > 1) {
	  return count;
	} else if (mode_mode==MODE_FILL && count > 0) {
	  return count;
	}
      } else {
	fprintf(stderr, _("Oops - conflict!\n"));
      }
    }
  }
  return count;
}

/* externally visible traverse function. See sudoku.h for the
   specification */
int sudoku_traverse(sudoku_t su, int mode, sudoku_callback_t *callback) {
  hypercube_t t;
  hypercube_t tr;
  int r;
  int count;

  /* translate from sudoku_t to hypercube_t */
  r = sudoku_to_hypercube(su, t);
  if (r == -1) {
    return 0;
  }
  count = traverse(t, r, mode, callback, 0, tr);
  if ((mode & MODE_MASK)==MODE_FILL) {
    /* translate back */
    hypercube_to_sudoku(tr, su);
  }
  return count;
}

static int visit(sudoku_t su, int visited[9][9], int i, int j, int x) {
  su[i][j] = 0;
  if (visited[i][j] != x) {
    visited[i][j] = x;
    return 1;
  }
  return 0;
}

static int visit_with_sym(int sym, sudoku_t su, int visited[9][9], int i, int j, int x) {
  int n=0;

  if (sym & 0x01) {
    n += visit(su, visited, i, j, x);
  }
  if (sym & 0x02) {
    n += visit(su, visited, i, 8-j, x);
  }
  if (sym & 0x04) {
    n += visit(su, visited, j, 8-i, x);
  }
  if (sym & 0x08) {
    n += visit(su, visited, 8-j, 8-i, x);
  }
  if (sym & 0x10) {
    n += visit(su, visited, 8-i, 8-j, x);
  }
  if (sym & 0x20) {
    n += visit(su, visited, 8-i, j, x);
  }
  if (sym & 0x40) {
    n += visit(su, visited, 8-j, i, x);
  }
  if (sym & 0x80) {
    n += visit(su, visited, j, i, x);
  }
  return n;
}

/* recude the given sudoku, which means, create as many blanks as
   possible while keeping the solution unique. sym is a symmetry
   from the above list. */
void sudoku_reduce(sudoku_t su, int sym) {
  sudoku_t su1;
  int visited[9][9];
  int n, i, j, r, count;

  /* remove pairs of points until solution becomes non-unique.
     Examine each pair of points exactly once, but in random
     order */
  memset(&visited, 0, sizeof(visited));
  n = 81; /* number of points not visited */
  while (n>0) {
    r = (rand() % n);
    for (i=0; i<9; i++) {
      for (j=0; j<9; j++) {
	if (!visited[i][j]) {
	  if (r==0) {
	    goto found;
	  }
	  r--;
	}
      }
    }
  found:
    /* remove a point, or several in case of symmetries */
    memcpy(su1, su, sizeof(sudoku_t));
    n -= visit_with_sym(sym, su1, visited, i, j, 1);

    count = sudoku_traverse(su1, MODE_UNIQUE, NULL);
    if (count == 1) {
      memcpy(su, su1, sizeof(sudoku_t));
    }
  }
}

#if 0
/* like sudoku_reduce, but try to maximize branches. Note: this is
   much slower than sudoku_reduce, but not much better, so we don't
   use it. It is left here for reference. */
static void sudoku_reduce_b(sudoku_t su, int sym) {
  sudoku_t su1;
  int visited[9][9];
  int m, n, i, j, r, count;
  int best_branches, best_i, best_j;

  /* remove pairs of points until solution becomes non-unique.
     Examine each pair of points exactly once, but in random
     order */
  memset(&visited, 0, sizeof(visited));
  m = 81; /* number of non-ones in visited array */
  while (m>0) {
    best_branches = 0;
    best_i = best_j = 0;
    n = m; /* number of zeros in visited array */
    while (n>0) {
      r = (rand() % n);
      for (i=0; i<9; i++) {
	for (j=0; j<9; j++) {
	  if (!visited[i][j]) {
	    if (r==0) {
	      goto found;
	    }
	    r--;
	  }
	}
      }
    found:
      /* remove a point, or several in case of symmetries */
      memcpy(su1, su, sizeof(sudoku_t));
      n -= visit_with_sym(sym, su1, visited, i, j, 2);

      count = sudoku_traverse(su1, MODE_UNIQUE, NULL);
      if (count > 1) {
	m -= visit_with_sym(sym, su1, visited, i, j, 1);
      } else {
	count = sudoku_traverse(su1, MODE_BRANCHES, NULL);
	if (count > best_branches) {
	  best_branches = count;
	  best_i = i;
	  best_j = j;
	}
      }
    }
    if (best_branches > 0) {
      m -= visit_with_sym(sym, su, visited, best_i, best_j, 1);
    }
    for (i=0; i<9; i++) {
      for (j=0; j<9; j++) {
	if (visited[i][j] == 2) {
	  visited[i][j] = 0;
	}
      }
    }
  }
}
#endif

/* create a sudoku puzzle according to the given parameters. Return
   puzzle in su, and the answer in ans. Return the number of branches
   in the search tree (a measure of difficulty). If pattern is given,
   find a sudoku fitting the pattern. If reduce==1, reduce the
   answer. (Either a pattern or reduce or both must be specified). Sym
   is a symmetry type of SYM_RANDOM. */
int sudoku(sudoku_t su, sudoku_t ans, int pattern[9][9], int reduce, int sym) {
  int count;
  int i, j;

  while (1) {
    /* create a random solution */
    memset(su, 0, sizeof(sudoku_t));
    sudoku_traverse(su, MODE_FILL | FLAG_RANDOM, NULL);
    memcpy(ans, su, sizeof(sudoku_t));

    if (pattern != NULL) {
      for (i=0; i<9; i++) {
	for (j=0; j<9; j++) {
	  if (!pattern[i][j]) {
	    su[i][j] = 0;
	  }
	}
      }

      count = sudoku_traverse(su, MODE_UNIQUE, NULL);
      if (count > 1) {
	continue;
      }
    }

    if (reduce) {
      sudoku_reduce(su, sym);
    }

    /* count the number of branches in the search tree */
    count = sudoku_traverse(su, MODE_BRANCHES, NULL);
    return count;
  }    
}

